import json
from typing import List

from kubetwo.model import CreateInput ,ScaleInput
from kubetwo.config import ArtifactSettings
from kubetwo.ansible import Ansible
from kubetwo.common import Common
from kubetwo.terraform import Terraform


class Scale:

    def __init__(self, input: ScaleInput):
        self.input = input
        self.artifact = ArtifactSettings(self.input.cluster_name)

    def post_action(self):
        Common.create_node_info(
            self.artifact.RENDERED_TFSTATE_PATH,
            self.input.ssh_private_key,
            self.artifact.RENDERED_NODE_INFO_PATH
        )
        Common.save_input_info(
            self.artifact.INPUT_JSON_FILE_PATH, 
            CreateInput(**self.input.dict())
        )


class ScaleIn(Scale):
    
    def run(self):
        terraform = Terraform(self.input.cluster_name)
        terraform.create_terraform_manifests(**self.input.dict())
        terraform.show_plan()
        if not self.input.approve:
            terraform.check_tf_plan()
        
        ansible = Ansible(
            self.input.cluster_name,
            Common.get_default_user_name(self.artifact.RENDERED_TFSTATE_PATH),
            self.input.ssh_private_key
        )
        ansible.run_kubespray_scale_in(self.get_remove_nodes())

        terraform.apply()
        ansible.create_inventory_file()

        self.post_action()

    def get_remove_nodes(self) -> List[str]:
        with open(self.artifact.INPUT_JSON_FILE_PATH, "r") as f:
            current_worker_node_num = json.load(f)["worker_node"]
        future_worker_node_num = self.input.worker_node

        remove_nodes = []
        for i in range(current_worker_node_num - future_worker_node_num):
            remove_nodes.append(f"worker-node{current_worker_node_num - i - 1}")
        return remove_nodes


class ScaleOut(Scale):
    
    def run(self):
        terraform = Terraform(self.input.cluster_name)
        terraform.create_terraform_manifests(**self.input.dict())
        terraform.show_plan()
        if not self.input.approve:
            terraform.check_tf_plan()
        terraform.apply()

        ansible = Ansible(
            self.input.cluster_name,
            Common.get_default_user_name(self.artifact.RENDERED_TFSTATE_PATH),
            self.input.ssh_private_key
        )
        ansible.create_inventory_file()
        ansible.run_kubespray_scale_out()
        self.post_action()



import os
import shutil

from termcolor import cprint

from kubetwo.model import DeleteInput
from kubetwo.config import ArtifactSettings
from kubetwo.terraform import Terraform
from kubetwo.exception import *


class Delete:

    def __init__(self, input: DeleteInput):
        self.input = input
        self.artifact = ArtifactSettings(self.input.cluster_name)
        self.terraform = Terraform(self.input.cluster_name)

    def run(self):
        self.terraform.show_plan(destroy_mode=True)
        if not self.input.approve:
            self.check_tf_plan()
        self.terraform.destroy()
        if not self.input.approve:
            self.check_delete_dir()
        self.delete_dir()

    def check_tf_plan(self):
        cprint(f"Newly created AWS resources will be removed.", "yellow")
        cprint("Do you proceed?\n", "yellow")
        try:
            choice = input("Enter a value [y/N]: ")
            if choice.lower() in ['y', 'yes']:
                return
        except KeyboardInterrupt:
            print()

        raise CheckDeniedException("[ERROR] Suspended to remove AWS resources.")

    def check_delete_dir(self):
        cprint(f"{self.artifact.ARTIFACT_DIR_PATH} will be deleted.", "yellow")
        cprint("Do you proceed?\n", "yellow")
        try:
            choice = input("Enter a value [y/N]: ")
            if choice.lower() in ['y', 'yes']:
                return
        except KeyboardInterrupt:
            print()
        
        raise CheckDeniedException("[ERROR] Suspended to delete workspace.")

    def delete_dir(self):
        if os.path.isdir(str(self.artifact.ARTIFACT_DIR_PATH)):
            shutil.rmtree(str(self.artifact.ARTIFACT_DIR_PATH))



import os
import shutil
import tarfile

import requests
from termcolor import cprint

from kubetwo.model import CreateInput
from kubetwo.common import Common
from kubetwo.config import ArtifactSettings, Settings
from kubetwo.ansible import Ansible
from kubetwo.exception import *
from kubetwo.terraform import Terraform


class Create:

    def __init__(self, input: CreateInput):
        self.input = input
        self.artifact = ArtifactSettings(self.input.cluster_name)
    
    def run(self):
        if not self.input.approve:
            self.check_clean_artifact_dir()
        self.clean_artifact_dir()
        self.download_kubespray()

        terraform = Terraform(self.input.cluster_name)
        terraform.create_terraform_manifests(**self.input.dict())
        terraform.initialize()
        terraform.show_plan()
        if not self.input.approve:
            terraform.check_tf_plan()
        terraform.apply()

        ansible = Ansible(
            self.input.cluster_name,
            Common.get_default_user_name(self.artifact.RENDERED_TFSTATE_PATH),
            self.input.ssh_private_key
        )
        ansible.create_inventory_file()
        ansible.copy_group_vars()
        ansible.run_kubespray_create()
        ansible.run_setup()
        ansible.run_fetch_admin_conf(
            artifact_dir_path = self.artifact.ARTIFACT_DIR_PATH,
            admin_conf_path = self.artifact.ADMIN_CONF_PATH,
            first_control_plane = Common.get_node_info(self.artifact.RENDERED_TFSTATE_PATH)[0][0]
        )
        
        Common.create_node_info(
            self.artifact.RENDERED_TFSTATE_PATH,
            self.input.ssh_private_key,
            self.artifact.RENDERED_NODE_INFO_PATH
        )
        Common.save_input_info(
            self.artifact.INPUT_JSON_FILE_PATH, 
            self.input
        )

    def check_clean_artifact_dir(self):
        if not os.path.isdir(str(self.artifact.ARTIFACT_DIR_PATH)):
            return

        cprint(f"Directory {str(self.artifact.ARTIFACT_DIR_PATH)} already exists.", "yellow")
        cprint("Do you use this directory to setup kubetwo?\n", "yellow")
        try:
            choice = input("Enter a value [y/N]: ")
            print()
            if choice.lower() in ['y', 'yes']:
                return
        except KeyboardInterrupt:
            print()

    def clean_artifact_dir(self):
        if not os.path.isdir(str(self.artifact.ARTIFACT_DIR_PATH)):
            os.makedirs(str(self.artifact.ARTIFACT_DIR_PATH))

        if os.path.isdir(str(self.artifact.RENDERED_ANSIBLE_DIR_PATH)):
            shutil.rmtree(str(self.artifact.RENDERED_ANSIBLE_DIR_PATH))
        os.makedirs(str(self.artifact.RENDERED_ANSIBLE_DIR_PATH))
            
        if not os.path.isdir(str(self.artifact.RENDERED_TERRAFORM_DIR_PATH)):
            os.makedirs(str(self.artifact.RENDERED_TERRAFORM_DIR_PATH))
            
        for tf_file_path in self.artifact.RENDERED_TERRAFORM_DIR_PATH.glob("**/*.tf"):
            os.remove(str(tf_file_path))
        
        if os.path.isfile(str(self.artifact.RENDERED_INVENTORY_PATH)):
            os.remove(str(self.artifact.RENDERED_INVENTORY_PATH))

        cprint(f"Finished to clean up {str(self.artifact.ARTIFACT_DIR_PATH)}\n")

    def download_kubespray(self):
        cprint("Downloading Kubespray...")
        if os.path.isdir(self.artifact.KUBESPRAY_DIR_PATH):
            return
        
        kubespray_content = requests.get(Settings.KUBESPRAY_ARCHIVE_URL).content

        with open(self.artifact.KUBESPRAY_ARCHIVE_PATH ,mode='wb') as file:
            file.write(kubespray_content)

        with tarfile.open(self.artifact.KUBESPRAY_ARCHIVE_PATH) as tar:
            tar.extractall(self.artifact.ARTIFACT_DIR_PATH)

        os.remove(str(self.artifact.KUBESPRAY_ARCHIVE_PATH))
        cprint("Finished to download Kubespray.\n")



import json
import os
from pathlib import Path
import subprocess
from typing import Dict, List, Tuple

from jinja2 import Template
from pydantic import BaseModel
from termcolor import cprint

from kubetwo.config import Settings
from kubetwo.exception import *


class Common:

    @classmethod
    def run_command(cls, command: str, cwd: str=None, stdout: bool=True):
        stdout = None if stdout else subprocess.DEVNULL
        try:
            process = subprocess.run(
                command, 
                shell=True, 
                stdout=stdout, 
                stderr=subprocess.PIPE,
                cwd=cwd
            )
            process.check_returncode()
        except subprocess.CalledProcessError:
            raise ProcessException(process.stderr.decode("utf8"))
        
    @classmethod
    def get_node_info(cls, rendered_tfstate_path: Path) -> Tuple[List[Dict], List[Dict]]:
        """Load node information on AWS from terraform.tfstate.
           Each item format is {"name": str, "private_ip": str, "public_ip": str}."""
        if not os.path.exists(str(rendered_tfstate_path)):
            raise TerraformStateNotFound(f"""
                [ERROR] terraform.tfstate doesn't exist at {str(rendered_tfstate_path)}
                It's necessary to execute kubetwo init command first""")

        with open(rendered_tfstate_path, "r") as file:
            tfstate_data = json.load(file)

        control_plane_private_ip_list = tfstate_data["outputs"]["control_plane_private_ip"]["value"]
        control_plane_public_ip_list = tfstate_data["outputs"]["control_plane_public_ip"]["value"]
        worker_node_private_ip_list = tfstate_data["outputs"]["worker_node_private_ip"]["value"]
        worker_node_public_ip_list = tfstate_data["outputs"]["worker_node_public_ip"]["value"]

        control_plane_list = []
        for i, (private_ip, public_ip) in enumerate(zip(control_plane_private_ip_list, control_plane_public_ip_list)):
            name = f"control-plane{i}"
            control_plane_list.append({"name": name, "private_ip": private_ip, "public_ip": public_ip})

        worker_node_list = []
        for i, (private_ip, public_ip) in enumerate(zip(worker_node_private_ip_list, worker_node_public_ip_list)):
            name = f"worker-node{i}"
            worker_node_list.append({"name": name, "private_ip": private_ip, "public_ip": public_ip})

        return control_plane_list, worker_node_list
    
    @classmethod
    def get_default_user_name(cls, rendered_tfstate_path: Path) -> str:
        with open(rendered_tfstate_path, "r") as f:
            tfstate_data = json.load(f)

        ami_name = tfstate_data["outputs"]["ami_name"]["value"]
        ami_description = tfstate_data["outputs"]["ami_description"]["value"]
        ami_location = tfstate_data["outputs"]["ami_location"]["value"]

        with open(Settings.DISTRO_INFO_FILE_PATH, "r") as file:
            distro_list = json.load(file)

        default_user_name = "ec2-user"
        for distro in distro_list:
            if  (
                distro["ami_name_keyword"] in ami_name or
                distro["ami_description_keyword"] in ami_description or
                distro["ami_location_keyword"] in ami_location
                ):
                distro_name = distro["distro_name"]
                cprint(f"{distro_name} is detected.")
                default_user_name = distro["user_name"]
                break
                
        cprint(f"User \"{default_user_name}\" will be used for EC2 default user.\n")
        return default_user_name

    @classmethod
    def create_node_info(cls, 
        rendered_tfstate_path: Path,
        ssh_private_key_path: Path,
        rendered_node_info_path: Path
    ):
        control_plane_list, worker_node_list = cls.get_node_info(rendered_tfstate_path)
        user_name = cls.get_default_user_name(rendered_tfstate_path)
        params = {
            "control_plane_list": control_plane_list,
            "worker_node_list": worker_node_list,
            "user_name": user_name,
            "ssh_private_key": str(ssh_private_key_path)
        }
        with open(Settings.NODE_INFO_FILE_PATH, "r") as file:
            template = Template(file.read())

        node_info = template.render(params)

        with open(rendered_node_info_path, "w") as file:
            file.write(node_info)

        cprint(node_info, "cyan")

    @classmethod
    def save_input_info(cls, input_json_path: Path, *inputs: BaseModel):
        merged_dict = {}
        for input in inputs:
            input_dict = json.loads(input.json())
            merged_dict = {**merged_dict, **input_dict}

        with open(input_json_path, "w") as f:
            json.dump(merged_dict, f, indent=4)

    @classmethod
    def load_input_info(cls, input_json_path: Path) -> Dict:
        pass



class ProcessException(Exception):
    pass

class CheckDeniedException(Exception):
    pass

class AnsibleException(ProcessException):
    pass

class AnsibleKubesprayException(AnsibleException):
    pass

class AnsibleSetupException(AnsibleException):
    pass

class AnsibleDeploySampleException(AnsibleException):
    pass

class AnsibleFetchAdminConfException(AnsibleException):
    pass

class TerraformException(ProcessException):
    pass

class TerraformInitException(TerraformException):
    pass

class TerraformPlanException(TerraformException):
    pass

class TerraformApplyException(TerraformException):
    pass

class TerraformDestroyException(TerraformException):
    pass

class TerraformStateNotFound(Exception):
    pass



import os
from pathlib import Path
from typing import List


class Settings:
    # Kubespray related
    KUBESPRAY_VERSION = "2.16.0"
    KUBESPRAY_ARCHIVE_URL = f"https://github.com/kubernetes-sigs/kubespray/archive/refs/tags/v{KUBESPRAY_VERSION}.tar.gz"

    # AWS related
    DEFAULT_AMI: str = "ami-0c3fd0f5d33134a76"
    DEFAULT_INSTANCE_TYPE: str = "t3.medium"
    DEFAULT_AVAILABILITY_ZONE: str = "us-west-1a"
    DEFAULT_CONTROL_PLANE_NUM: int = 1
    DEFAULT_WORKER_NODE_NUM: int = 1
    DEFAULT_OPEN_PORTS: List[int] = [-1]

    # Directory names
    TERRAFORM_TEMPLATE_DIR_NAME: str = "terraform_template"
    ANSIBLE_TEMPLATE_DIR_NAME: str = "ansible_template"
    KUBERNETES_SAMPLE_DIR_NAME: str = "kubernetes_sample"
    DATA_DIR_NAME: str = "data"
    
    # File names
    INVENTORY_FILE_NAME: str = "inventory.yml"
    SETUP_PLAYBOOK_NAME: str = "setup.yml"
    FETCH_ADMIN_CONF_NAME: str = "fetch_admin_conf.yml"
    DISTRO_INFO_FILE_NAME: str = "distro_info.json"
    NODE_INFO_FILE_NAME: str = "node_info.txt.j2"
    VALIDATION_RESULT_FILE_NAME: str = "validation_result.txt.j2"

    # Directory paths
    SELF_PATH: Path = Path(os.path.dirname(__file__)) # Absolute path of the directory where this file belongs to
    TERRAFORM_TEMPLATE_DIR_PATH: Path = SELF_PATH / TERRAFORM_TEMPLATE_DIR_NAME
    ANSIBLE_TEMPLATE_DIR_PATH: Path = SELF_PATH / ANSIBLE_TEMPLATE_DIR_NAME
    KUBERNETES_SAMPLE_DIR_PATH: Path = SELF_PATH / KUBERNETES_SAMPLE_DIR_NAME
    DATA_DIR_PATH: Path = SELF_PATH / DATA_DIR_NAME

    # File paths
    INVENTORY_FILE_PATH: Path = ANSIBLE_TEMPLATE_DIR_PATH / INVENTORY_FILE_NAME
    SETUP_PLAYBOOK_PATH: Path = ANSIBLE_TEMPLATE_DIR_PATH / SETUP_PLAYBOOK_NAME
    FETCH_ADMIN_CONF_PATH: Path = ANSIBLE_TEMPLATE_DIR_PATH / FETCH_ADMIN_CONF_NAME
    DISTRO_INFO_FILE_PATH: Path = DATA_DIR_PATH / DISTRO_INFO_FILE_NAME
    NODE_INFO_FILE_PATH: Path = DATA_DIR_PATH / NODE_INFO_FILE_NAME
    VALIDATION_RESULT_FILE_PATH: Path = DATA_DIR_PATH / VALIDATION_RESULT_FILE_NAME



class ArtifactSettings:

    # Directory names (Artifact)
    KUBESPRAY_ARCHIVE_NAME: str = f"v{Settings.KUBESPRAY_VERSION}.tar.gz"
    KUBESPRAY_DIR_NAME: str = f"kubespray-{Settings.KUBESPRAY_VERSION}"
    RENDERED_TERRAFORM_DIR_NAME: str = "terraform"
    RENDERED_ANSIBLE_DIR_NAME: str = "ansible"
    RENDERED_KUBERNETES_DIR_NAME: str = "kubernetes"
    RENDERED_INVENTORY_NAME: str = "inventory.yml"
    RENDERED_NODE_INFO_NAME: str = "node_info.txt"
    INPUT_JSON_FILE_NAME: str = "input.json"

    def __init__(self, cluster_name: str):
        # Directory paths (Artifact)
        self.ARTIFACT_DIR_PATH: Path = Path.cwd() / cluster_name.replace('-', '_')
        self.KUBESPRAY_ARCHIVE_PATH: Path = self.ARTIFACT_DIR_PATH / self.KUBESPRAY_ARCHIVE_NAME
        self.KUBESPRAY_DIR_PATH: Path = self.ARTIFACT_DIR_PATH / self.KUBESPRAY_DIR_NAME
        self.KUBESPRAY_GROUP_VARS_DIR_PATH: Path = self.KUBESPRAY_DIR_PATH / "inventory/sample/group_vars"
        self.RENDERED_TERRAFORM_DIR_PATH: Path = self.ARTIFACT_DIR_PATH / self.RENDERED_TERRAFORM_DIR_NAME
        self.RENDERED_ANSIBLE_DIR_PATH: Path = self.ARTIFACT_DIR_PATH / self.RENDERED_ANSIBLE_DIR_NAME
        self.RENDERED_KUBERNETES_DIR_PATH: Path = self.ARTIFACT_DIR_PATH / self.RENDERED_KUBERNETES_DIR_NAME
        self.RENDERED_TFSTATE_PATH: Path = self.RENDERED_TERRAFORM_DIR_PATH / "terraform.tfstate"
        self.RENDERED_INVENTORY_PATH: Path = self.RENDERED_ANSIBLE_DIR_PATH / self.RENDERED_INVENTORY_NAME
        self.RENDERED_KUBESPRAY_GROUP_VARS_DIR_PATH: Path = self.RENDERED_ANSIBLE_DIR_PATH / "group_vars"
        self.RENDERED_NODE_INFO_PATH: Path = self.ARTIFACT_DIR_PATH / self.RENDERED_NODE_INFO_NAME
        self.ADMIN_CONF_PATH: Path = self.ARTIFACT_DIR_PATH / "admin.conf"
        self.INPUT_JSON_FILE_PATH: Path = self.ARTIFACT_DIR_PATH / self.INPUT_JSON_FILE_NAME



import os
from codecs import open
from pathlib import Path
from typing import List

from setuptools import setup


def package_files(directory: str, excelude_suffix_list: List=[]) -> List[str]:
    directory_path = Path("kubetwo") / directory
    paths = []
    excelude_suffix_list = ["." + excelude_suffix for excelude_suffix in excelude_suffix_list]
    for path in directory_path.glob("**/*"):
        if path.suffix in excelude_suffix_list:
            continue
        paths.append(str(path.relative_to("kubetwo")))
    return paths


all_package_files = \
    package_files("ansible_template") + \
    package_files("data") + \
    package_files("kubernetes_sample") + \
    package_files("terraform_template")

here = Path(os.path.dirname(__file__))

with open(here / "README.md", encoding="utf-8") as f:
    long_description = f.read()

with open(here / "requirements.txt") as f:
    install_requires = f.read().splitlines()

setup(
    name="kubetwo",
    version="0.2.0",
    description="Simple CLI tool to create Kubernetes cluster on AWS EC2.",
    long_description=long_description,
    author="opeco17",
    url="https://github.com/opeco17/kubetwo",
    long_description_content_type="text/markdown",
    packages=["kubetwo"],
    package_data={"kubetwo": all_package_files},
    install_requires=install_requires,
    license="Apache License 2.0",
    python_requires=">= 3.6",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology"
    ],
    keywords="kubetwo kube-two kube2 kube-2",
    entry_points={
        "console_scripts": [
            "kubetwo = kubetwo.cli:main"
        ]
    }
)
